import math
import torch
import torch.nn as nn
from itertools import chain
from torch.nn.utils.rnn import pad_packed_sequence, pack_padded_sequence

from .encoder import EncoderBase
from ..utils import log_sum_exp
from sympy import *

def mocoloss(x):
    kernelmatrix = torch.mm(x, x.T)
    expofkernel = torch.exp(kernelmatrix)
    realcross = torch.diagonal(expofkernel)[:x.size()[0]//2]
    sum_inners = torch.sum(expofkernel,dim=1)
    realcross = torch.div(realcross, sum_inners[:x.size()[0]//2])
    return -torch.sum(torch.log(realcross))

class EPositionalEncoding(nn.Module):
    r"""Inject some information about the relative or absolute position of the tokens
        in the sequence. The positional encodings have the same dimension as
        the embeddings, so that the two can be summed. Here, we use sine and cosine
        functions of different frequencies.
    .. math::
        \text{PosEncoder}(pos, 2i) = sin(pos/10000^(2i/d_model))
        \text{PosEncoder}(pos, 2i+1) = cos(pos/10000^(2i/d_model))
        \text{where pos is the word position and i is the embed idx)
    Args:
        d_model: the embed dim (required).
        dropout: the dropout value (default=0.1).
        max_len: the max. length of the incoming sequence (default=5000).
    Examples:
        >>> pos_encoder = PositionalEncoding(d_model)
    """

    def __init__(self, d_model, dropout=0.1, max_len=400):
        super(EPositionalEncoding, self).__init__()
        self.dropout = nn.Dropout(p=dropout)

        pe = torch.zeros(max_len, d_model)
        position = torch.arange(0, max_len, dtype=torch.float).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2).float() * (-math.log(10000.0) / d_model))
        pe[:, 0::2] = torch.sin(position * div_term)
        pe[:, 1::2] = torch.cos(position * div_term)
        pe = pe.unsqueeze(0).transpose(0, 1)
        self.register_buffer('pe', pe)
        self.ninp = d_model

    def forward(self, x):
        r"""Inputs of forward function
        Args:
            x: the sequence fed to the positional encoder model (required).
        Shape:
            x: [sequence length, batch size, embed dim]
            output: [sequence length, batch size, embed dim]
        Examples:
            >>> output = pos_encoder(x)
        """

        x = x*math.sqrt(self.ninp) + self.pe[:x.size(0), :]
        return self.dropout(x)

class etransposer(nn.Module):
    def __init__(self):
        super(etransposer, self).__init__()

    def forward(self,x):
        return (torch.transpose(x,0,1))

class TransEncoderBase(EncoderBase):
    """docstring for EncoderBase"""
    def __init__(self):
        super(TransEncoderBase, self).__init__()

    def freeze(self):
        for param in self.parameters():
            param.requires_grad = False

    def forward(self, x):
        """
        Args:
            x: (batch_size, *)

        Returns: Tensor1, Tensor2
            Tensor1: the mean tensor, shape (batch, nz)
            Tensor2: the logvar tensor, shape (batch, nz)
        """

        raise NotImplementedError

    def encode_stats(self, x):

        return self.forward(x)

    def sample(self, input, nsamples):
        """sampling from the encoder
        Returns: Tensor1
            Tensor1: the tensor latent z with shape [batch, nsamples, nz]
        """

        # (batch_size, nz)
        mu, logvar = self.forward(input)

        # (batch, nsamples, nz)
        z = self.reparameterize(mu, logvar, nsamples)

        return z, (mu, logvar)

    def encode(self, input, nsamples, ghostvae=None):
        """perform the encoding and compute the KL term

        Returns: Tensor1, Tensor2
            Tensor1: the tensor latent z with shape [batch, nsamples, nz]
            Tensor2: the tenor of KL for each x with shape [batch]

        """

        # (batch_size, nz)
        mu, logvar = self.forward(input)

        #print(mu.size())

        #mu = mu.view(-1,mu.shape[-1])
        #logvar = logvar.view(-1,logvar.shape[-1])

        if(ghostvae is not None):
            ghostsamples, _ = ghostvae.encoder.forward(input) 
            #print(ghostsamples.size())
            totalsamples = torch.cat([mu,ghostsamples],dim=0)
            KL = mocoloss(totalsamples)
        else:
            KL = 0


        batch_size, nz = mu.size()
        # (batch, nsamples, nz)
        #z = self.reparameterize(mu, logvar, nsamples)

        #print(z.size())

        #KL = 0.5 * (mu.pow(2) + logvar.exp() - logvar - 1).sum(dim=1)

        #return z, KL
        return mu.unsqueeze(1).expand(batch_size, nsamples, nz), KL

    def reparameterize(self, mu, logvar, nsamples=1):
        """sample from posterior Gaussian family
        Args:
            mu: Tensor
                Mean of gaussian distribution with shape (batch, nz)

            logvar: Tensor
                logvar of gaussian distibution with shape (batch, nz)

        Returns: Tensor
            Sampled z with shape (batch, nsamples, nz)
        """
        #mu = mu.view(-1,mu.shape[-1])
        #logvar = logvar.view(-1,logvar.shape[-1])
        batch_size, nz = mu.size()
        std = logvar.mul(0.5).exp()

        mu_expd = mu.unsqueeze(1).expand(batch_size, nsamples, nz)
        std_expd = std.unsqueeze(1).expand(batch_size, nsamples, nz)

        eps = torch.zeros_like(std_expd).normal_()

        return mu_expd + torch.mul(eps, std_expd)

    def eval_inference_dist(self, x, z, param=None):
        """this function computes log q(z | x)
        Args:
            z: tensor
                different z points that will be evaluated, with
                shape [batch, nsamples, nz]
        Returns: Tensor1
            Tensor1: log q(z|x) with shape [batch, nsamples]
        """

        nz = z.size(2)

        if not param:
            mu, logvar = self.forward(x)
        else:
            mu, logvar = param

        # (batch_size, 1, nz)
        mu, logvar = mu.unsqueeze(1), logvar.unsqueeze(1)
        var = logvar.exp()

        # (batch_size, nsamples, nz)
        dev = z - mu

        # (batch_size, nsamples)
        log_density = -0.5 * ((dev ** 2) / var).sum(dim=-1) - \
            0.5 * (nz * math.log(2 * math.pi) + logvar.sum(-1))

        return log_density

    def calc_mi(self, x):
        """Approximate the mutual information between x and z
        I(x, z) = E_xE_{q(z|x)}log(q(z|x)) - E_xE_{q(z|x)}log(q(z))

        Returns: Float

        """

        # [x_batch, nz]
        mu, logvar = self.forward(x)

        x_batch, nz = mu.size()

        # E_{q(z|x)}log(q(z|x)) = -0.5*nz*log(2*\pi) - 0.5*(1+logvar).sum(-1)
        neg_entropy = (-0.5 * nz * math.log(2 * math.pi)- 0.5 * (1 + logvar).sum(-1)).mean()

        # [z_batch, 1, nz]
        z_samples = self.reparameterize(mu, logvar, 1)

        # [1, x_batch, nz]
        mu, logvar = mu.unsqueeze(0), logvar.unsqueeze(0)
        var = logvar.exp()

        # (z_batch, x_batch, nz)
        dev = z_samples - mu

        # (z_batch, x_batch)
        log_density = -0.5 * ((dev ** 2) / var).sum(dim=-1) - \
            0.5 * (nz * math.log(2 * math.pi) + logvar.sum(-1))

        # log q(z): aggregate posterior
        # [z_batch]
        log_qz = log_sum_exp(log_density, dim=1) - math.log(x_batch)

        return (neg_entropy - log_qz.mean(-1)).item()

class GaussianTransEncoder(TransEncoderBase):
    """Gaussian LSTM Encoder with constant-length input"""
    def __init__(self, args, vocab_size, model_init, emb_init):
        super(GaussianTransEncoder, self).__init__()
        self.ni = args.ni
        self.nh = args.enc_nh
        self.nz = args.nz
        self.args = args

        self.embed = nn.Embedding(vocab_size, args.ni)

        #self.lstm = nn.LSTM(input_size=args.ni,
        #                    hidden_size=args.enc_nh,
        #                    num_layers=1,
        #                    batch_first=True,
        #                    dropout=0)

        self.encoder_layer = nn.TransformerEncoderLayer(d_model=args.ni, nhead=8, dim_feedforward=args.enc_nh)
        #self.transformer = nn.TransformerEncoder(self.encoder_layer, 6)
        self.transposer = etransposer()
        self.transformer = nn.TransformerEncoder(self.encoder_layer, 6)

        self.posenc = EPositionalEncoding(args.ni)
        
        # Just make embedding dim the output dim of the encoder before becoming z I guess

        self.linear = nn.Linear(args.ni, 2 * args.nz, bias=False)
        self.mu_bn = nn.BatchNorm1d(args.nz, affine=False)
        #self.mu_bn.weight.requires_grad = False

        self.reset_parameters(model_init, emb_init)

    def transform(self, src, mask=None, src_key_padding_mask=None):
        return self.transposer(self.transformer(self.transposer(src),mask,src_key_padding_mask))

    def postenc(self,x):
        return self.transposer(self.posenc(self.transposer(x)))

    def reset_parameters(self, model_init, emb_init, reset=False):
        # for name, param in self.lstm.named_parameters():
        #     # self.initializer(param)
        #     if 'bias' in name:
        #         nn.init.constant_(param, 0.0)
        #         # model_init(param)
        #     elif 'weight' in name:
        #         model_init(param)

        # model_init(self.linear.weight)
        # emb_init(self.embed.weight)
        if not reset:
            #if self.args.gamma > -1:
                #for param in self.parameters():
                #    model_init(param)
                #emb_init(self.embed.weight)
            #self.mu_bn.weight.fill_(self.args.gamma)
            print("going on..")
        else:
            print('reset bn!')
            self.mu_bn.weight.fill_(self.args.gamma)
            nn.init.constant_(self.mu_bn.bias, 0.0)

    # this part needs to have the mask tensor

    def forward(self, input):
        """
        Args:
            x: (batch_size, seq_len)

        Returns: Tensor1, Tensor2
            Tensor1: the mean tensor, shape (batch, nz)
            Tensor2: the logvar tensor, shape (batch, nz)
        """

        # (batch_size, seq_len-1, args.ni)
        word_embed = self.embed(input)

        #_, (last_state, last_cell) = self.lstm(word_embed)

        word_embed = self.postenc(word_embed)

        rep = self.transform(word_embed)

        rep_relevant = rep[:,0,:]

        mean,logvar = self.linear(rep_relevant).chunk(2,-1)

        #print(mean.size())

        #mean, logvar = self.linear(last_state).chunk(2, -1)
        if self.args.gamma >= 0:
            #mean = self.mu_bn(mean)
            mean = self.mu_bn(mean)
        else:
            mean = mean
        # fix variance as a pre-defined value
        if self.args.fix_var > 0:
            logvar = mean.new_tensor([[[math.log(self.args.fix_var)]]]).expand_as(mean)
            
        return mean, logvar